;;;; ss-mode.el -- Control Visual Source Safe version control system from Emacs

;; Author:        Steve Molitor

;; July 25, 1999


;;; Commentary

;; Similar to vc-mode, but not as complete.

;; Since we don't use version control headers where I work, ss-mode
;; uses shell commands to query the status of a file.  This can be
;; slow.  To avoid freezing up all of Emacs while waiting for a Source
;; Safe shell command to return, I use asynchronous processes.  This
;; makes the implementation a little more complicated.
;;
;; Also, since we don't use version control comments where I work (bad
;; bad bad!), I have not implemented entering in comments.  This needs
;; to be added.
;;
;; To view online help on ss-mode, run the command ss-mode-help.



;;; Code

;; Customization variables ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defgroup ss nil
  "*Version control with Visual Source Safe."
  :group 'my-stuff)

(defcustom ss-user-name ""
  "*Source Safe user name.
If empty, will default to value of environment variable USERNAME."
  :type 'string
  :group 'ss)

(defcustom ss-file-extensions-list
  '(".h"
    ".hpp"
    ".c"
    ".cpp"
    ".ec"
    ".cmd"
    ".cat"
    ".bat"
    ".pl"
    ".pm")
  "*List of file extensions for files that can be in Source Safe.
If the list is nil, file extensions are not checked by ss-mode before
querying Source Safe for the status of a file."
  :type 'sexp
  :group 'ss)

(defcustom ss-directories-list
  '("d:\\\\enterprise")
  "*List of directories that can contain Source Safe files.
If the list is nil, ss-mode does not check the directory of a file before
querying Source Safe for the status of a file."
  :type 'sexp
  :group 'ss)

(defcustom ss-name "SS"
  "*Name for Source Safe appearing on the mode line."
  :type 'string
  :group 'ss)

(defcustom ss-checked-in-symbol "-"
  "*Symbol that appears on mode line indicating that file is checked in."
  :type 'string
  :group 'ss)

(defcustom ss-checked-out-symbol ":"
  "*Symbol that appears on the mode line indicating that file is checked in."
  :type 'string
  :group 'ss)

(defcustom ss-diff-format-args "-D"
  "*Format arguments to Source Safe diff command."
  :type 'string
  :group 'ss)

(defcustom ss-update-interval 30
  "*Update Source Safe status of current buffer this number of
seconds, when Emacs is idle"
  :type 'integer
  :group 'ss)

(defcustom ss-db-list (list (list "Edge" "\\\\srclib1\\vss_fltsys\\edge")
			    (list "Architecture" "\\\\srclib1\\vss_fltsys\\Architecture"))
  "List of Source Safe databases.

Each element of the list a list of two elements.  The first element is
the name of the database to be displayed on the menu.  The second
element is the database path.

The first database in the list will be the default."
  :type 'sexp
  :group 'ss)


(defvar ss-prefix "\C-cv"
  "*Prefix key for ss-mode commands.")
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;




;; Setup this minor mode ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defvar ss-mode nil
  "Mode-line variable for ss-mode.")
(make-variable-buffer-local 'ss-mode)

(if (not (assq 'ss-mode minor-mode-alist))
    (setq minor-mode-alist (cons '(ss-mode ss-mode) minor-mode-alist)))

;; Keymap
;; Bindings for this have to go in the global map, as we'll often
;; want to call them from random buffers.
(defvar ss-keymap nil)
(setq ss-keymap (lookup-key global-map ss-prefix))
(if (not ss-keymap)
    (progn
      (setq ss-keymap (make-sparse-keymap))
      (define-key global-map ss-prefix ss-keymap)
      (define-key ss-keymap "u" 'ss-undo-checkout)
      (define-key ss-keymap "s" 'ss-status)
      (define-key ss-keymap "i" 'ss-register)
      (define-key ss-keymap "l" 'ss-print-log)
      (define-key ss-keymap "v" 'ss-next-action)
      (define-key ss-keymap "p" 'ss-get-project)
      (define-key ss-keymap "g" 'ss-get)
      (define-key ss-keymap "m" 'ss-files-checked-out-by-me)
      (define-key ss-keymap "d" 'ss-directory)
      (define-key ss-keymap "=" 'ss-diff)
      (define-key ss-keymap "e" 'ss-delete)
      ))      


;; menus -- xemacs only
(if (featurep 'xemacs)
    (progn
      (add-submenu nil '("Source Safe"))
      (add-menu-button '("Source Safe") [ "Refresh Source Safe Status" ss-status ])
      (add-menu-button '("Source Safe") [ "Register" ss-register ])
      (add-menu-button '("Source Safe") [ "History" ss-print-log ])
      (add-menu-button '("Source Safe") [ "Next Action" ss-next-action ])
      (add-menu-button '("Source Safe") [ "Get Project" ss-get-project ])
      (add-menu-button '("Source Safe") [ "Get" ss-get ])
      (add-menu-button '("Source Safe") [ "Files checked out by me" ss-files-checked-out-by-me ])
      (add-menu-button '("Source Safe") [ "Directory" ss-directory ])
      (add-menu-button '("Source Safe") [ "Diff" ss-diff ])
      (add-menu-button '("Source Safe") [ "Delete" ss-delete ])
      (add-submenu '("Source Safe") '("Select SS Database" :filter ss-db-menu-filter))
      (add-menu-button '("Source Safe") "-----")
      (add-menu-button '("Source Safe") [ "Source Safe Help" ss-mode-help ])))


;; set default ss database
(setenv "SSDIR" (car (cdr (car ss-db-list))))



;; hooks, etc.
(add-hook 'find-file-hooks 'ss-status-hook)
(or (= ss-update-interval 0)
    (run-with-idle-timer ss-update-interval t 'ss-status-hook))

;; help
(defun ss-mode-help ()
  "Help for ss-mode."
  (interactive)
  (describe-function 'ss-mode))

(defvar ss-prefix-keymap nil)
(setq ss-prefix-keymap (lookup-key global-map ss-prefix))

(defun ss-mode ()
  "Ss-mode is a minor mode for controlling Visual Source Safe from Emacs.


COMMANDS:

The prefix key is controlled by the value of ss-prefix, which you can
set in your .emacs file.  By default it is \"C-cv\".

\\{ss-keymap}")
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;




;; Asynchronous process support ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; We allow one Source Safe process to be running concurrently per
;; buffer.  Functions can place tasks (Source Safe commands) to be
;; executed on the ss-tasks-queue, and then run all of the enqueued
;; tasks sequentially with ss-run-tasks.


(defvar ss-proc-alist nil
  "Mapping between process an buffer.")

(defvar ss-queue nil)
(make-variable-buffer-local 'ss-queue)
(set-default 'ss-queue nil)

(defvar ss-running nil)
(make-variable-buffer-local 'ss-running)
(set-default 'ss-running nil)


(defun ss-enqueue (func prog &optional i-func &rest args)
  "Put task on queue.

First argument FUNC is the function that will be called when the
process is finished to interpret the process output.  This function
should take two arguments.  The first is the process exit status (an
integer), the second is the buffer containing the output of the
process.

Second argument PROG is the program to be executed.

Third argument I-FUNC is a optional function used to send input to the
process.  If I-FUNC is non-nil, we call that function and send it's
result to the process.  I-FUNC should return a string.

Fourth argument ARGS is a list of the arguments passed to the program.

The current buffer active when ss-enqueue is invoked is saved, and
restored when the task is processeed, so that buffer-local variables
are available to the parse function FUNC."
  (setq ss-queue (append ss-queue (list (list
					 'parse-func func
					 'program prog
                                         'input-func i-func
					 'program-args args
					 'user-buf (current-buffer))))))

(defun ss-run-tasks ()
  "Process all of the tasks on current buffer's tasks queue."
  (if ss-running
      (error "Another Source Safe process is running for this buffer."))
  (setq ss-running t)
  (ss-do-task))
    

(defun ss-do-task ()
  "Process one task off of ss-queue.

The sentinel function will recursively call ss-do-task until the queue
is empty, so calling ss-do-task will process all of the tasks on the
queue.

This function is used by ss-run-tasks and ss-task-finished.  Do not
call it directly; use ss-run-tasks instead."
  (if ss-queue
      (progn
	(let* ((task (car ss-queue))
               (buf (generate-new-buffer "*Source Safe Output*"))
	       (proc (apply 'start-process
			    "ss-process"
			    buf
			    (plist-get task 'program)
			    (plist-get task 'program-args))))
          (process-kill-without-query proc)
	  (let ((i-func (plist-get task 'input-func)))
	    (if i-func
	  	(process-send-string proc (funcall i-func))))
	  (setq ss-proc-alist (cons (list proc (current-buffer)) ss-proc-alist))
	  (set-process-sentinel proc 'ss-task-finished)))
    (setq ss-running nil)))


(defun ss-task-finished (proc event)
  "Sentinel function for Source Safe processes."
  (let* ((proc-info (assq proc ss-proc-alist))
	 (u-buf (car (cdr proc-info)))
	 (task (car ss-queue)))
    (unwind-protect
	(progn
	  (setq ss-proc-alist (delq proc-info ss-proc-alist))
	  (setq ss-queue (cdr ss-queue))
	  (if (buffer-live-p u-buf)
	      (progn
		(set-buffer u-buf)
		(condition-case err
		    (funcall (plist-get task 'parse-func)
			     (process-exit-status proc)
			     (process-buffer proc))
		  (error
		   (setq ss-queue nil)
		   (ss-display-output (process-buffer proc) "*Source Safe Error*")
		   (message (princ (format "Error: %s" err)))
		   (ding)                
		   (setq ss-queue nil))))))
      (kill-buffer (process-buffer proc))
      (ss-do-task))))


(defun ss-clear ()
  "Clear running Source Safe tasks."
  (interactive)
  (setq ss-queue nil)
  (setq ss-running nil))
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;




;; Source Safe Status ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defvar ss-status nil)
(make-variable-buffer-local 'ss-status)
(set-default 'ss-status nil)


(defun ss-status ()
  "Update Source Safe status of current buffer."
  (interactive)
  (ss-buffer-file-check)
  (ss-enqueue 'ss-parse-status "ss" nil "status" "-I-" (ss-path buffer-file-name))
  (ss-run-tasks))


(defun ss-parse-status (rc ss-output-buf)
  "Parse results of Source Safe status query."
  (let (st
	co-by)
    (cond ((equal rc 0)		; checked in
	    (setq st 'checked-in))
	   ((equal rc 1)	; checked out
	    (setq st 'checked-out)
	    (save-excursion
	      (set-buffer ss-output-buf)
	      (goto-char (point-min))
	      (setq co-by
		    (buffer-substring (re-search-forward "\\s-+") (re-search-forward "\\>")))))
	   ((equal rc 100)	; not in source safe
	    (setq st 'not-in-ss))
	   (t			; error
	    (error "Status operation failed")))
    (setq ss-status (list 'state st 'checked-out-by co-by)))
  (ss-update-mode-line))


(defun ss-update-mode-line ()
  "Update mode-line to display Source Safe status."
  (cond ((eq (plist-get ss-status 'state) 'checked-in)
	  (setq ss-mode (concat " " ss-name ss-checked-in-symbol)))
	 ((eq (plist-get ss-status 'state) 'checked-out)
	  (setq ss-mode (concat " " ss-name ss-checked-out-symbol))
	  (or (string= (downcase (plist-get ss-status 'checked-out-by)) (downcase (ss-user-name)))
	      (setq ss-mode (concat ss-mode (plist-get ss-status 'checked-out-by)))))
	 ((eq (plist-get ss-status 'state) 'not-in-ss)
	  (setq ss-mode nil))
	 (t
	  (error "Unexpected ss-status value")))
  (force-mode-line-update))
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;




; Source Safe Operations ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun ss-next-action ()
  "Peform next logical action on current buffer's file.

If the file is not registered with Source Safe, register it.
If the file is checked in, check it out.
If the file is checked out (by ss-user-name), check it in."
  (interactive)
  (ss-buffer-file-check)
  (if ss-running
      (error "Another Source Safe operation is running for this buffer"))
  (let ((st (plist-get ss-status 'state))
	 (co-by (plist-get ss-status 'checked-out-by)))
    (cond ((eq st 'checked-in)
           (ss-checkout))
          ((eq st 'checked-out)
           (if (equal (downcase co-by) (downcase (ss-user-name)))
               (ss-checkin)
             (error (concat (file-name-nondirectory buffer-file-name) " is checked out by " co-by))))
	   (t
	    (ss-register)))))



(defun ss-checkout ()
  "Check file out of Source Safe."
  (ss-buffer-file-check)
  (ss-enqueue 'ss-parse-checkout "ss" nil "checkout" "-I-" (ss-path buffer-file-name))
  (ss-run-tasks))
  
(defun ss-parse-checkout (rc ss-output-buffer)
  "Parse results of Source Safe checkout process."
  (if (equal rc 0)
      (progn
	    (ss-internal-revert-buffer)
	    (setq ss-status (list 'state 'checked-out 'checked-out-by (ss-user-name)))
	    (ss-update-mode-line))
    (error "Checkout operation failed")))



(defun ss-checkin ()
  "Check file in to Source Safe."
  (ss-buffer-file-check)
  (ss-enqueue 'ss-parse-checkin "ss" nil "checkin" "-I-" "-Ccomment" (ss-path buffer-file-name))
  (ss-run-tasks))

(defun ss-parse-checkin (rc ss-output-buf)
  "Parse results of Source Safe checkin process."
  (if (equal rc 0)
      (progn
	   (setq buffer-read-only t)
	   (setq ss-status (list 'state 'checked-in))
	   (ss-update-mode-line))
    (error "Checkin operation failed")))



(defun ss-register ()
  "Register file with current project.
The current project is determined by the directory of the
buffer file."
  (interactive)
  (ss-buffer-file-check)
  (ss-enqueue 'ss-parse-cp "ss" nil "cp" "-I-" (ss-path (file-name-directory buffer-file-name)))
  (ss-enqueue 'ss-parse-register "ss" nil "add" "-I-" "-K" "-Ccomment"
              (file-name-nondirectory buffer-file-name))
  (ss-run-tasks))

(defun ss-parse-register (rc ss-output-buffer)
  (if (equal rc 0)
      (progn
	  (setq ss-status (list 'state 'checked-out 'checked-out-by (ss-user-name)))
	  (ss-update-mode-line))
    (error "Register operation failed")))

(defun ss-parse-cp (rc ss-output-buffer)
  (if (not (equal rc 0))
      (error "Change project operation failed")))


(defun ss-delete ()
  "Delete file from Source Safe."
  (interactive)
  (ss-buffer-file-check)
  (if (y-or-n-p (concat "Really delete " (file-name-nondirectory buffer-file-name) " from Source Safe? "))
      (progn
	(message "")
	(ss-enqueue 'ss-parse-delete "ss" nil "delete" "-I-" (ss-path buffer-file-name))
	(ss-run-tasks))))

(defun ss-parse-delete (rc ss-output-buffer)
  (if (equal rc 0)
      (progn
	    (setq ss-status (list 'state 'not-in-ss))
	    (ss-update-mode-line))
    (error "Delete operation failed")))



(defun ss-get ()
  "Get latest version of file from Source Safe."
  (interactive)
  (ss-buffer-file-check)
  (if (not ss-mode)
      (error "%s%s" (file-name-nondirectory buffer-file-name) " is not registered with Source Safe."))
  (ss-enqueue 'ss-parse-get "ss" nil "get" "-I-" (ss-path buffer-file-name))
  (ss-run-tasks))

(defun ss-parse-get (rc ss-output-buf)
  (if (equal rc 0)
      (let (msg)
	  (save-excursion
	    (set-buffer ss-output-buf)
	    (goto-char (point-min))
	    (if (re-search-forward "eplacing" nil t)
		(progn
		  (ss-internal-revert-buffer)
		  (setq msg "Retrieved latest version of "))
	      (setq msg "No change to ")))
	  (message "%s%s" msg buffer-file-name))
    (error "%s%s" "Get operation failed")))



(defun ss-get-project (project)
  "Recursively retrieve latest version of project tree from Source Safe."
  (interactive "DProject: ")
  (if project
      (progn
	   (setq project (ss-path project))
	   (ss-enqueue 'ss-parse-get-project "ss" nil "get" project "-I-" "-R")
	   (ss-run-tasks))))

(defun ss-parse-get-project (rc ss-output-buf)
  (if (equal rc 0)
      (ss-display-output ss-output-buf "*Source Safe Get Results*")
    (error "Get Project operation failed")))



(defun ss-print-log ()
  "Show history of current buffer in a window."
  (interactive)
  (ss-buffer-file-check)
  (ss-enqueue 'ss-parse-print-log "ss" nil "history" "-I-" (ss-path buffer-file-name))
  (ss-run-tasks))

(defun ss-parse-print-log (rc ss-output-buf)
  (if (equal rc 0)
      (ss-display-output ss-output-buf "*Source Safe History*")
    (error "History operation failed")))



(defun ss-diff (arg)
  "Display difference between two Source Safe files.

By default, display differen between local copy of current buffer's
file and Source Safe's copy.

With prefix argument, prompt for version to compare local file to."
  (interactive "P")
  (ss-buffer-file-check)
  (if arg
      (let ((version (read-string "Version: ")))
	(message "")
	(ss-enqueue 'ss-parse-diff "ss" nil "diff" "-I-"
		    (concat "-V" version) ss-diff-format-args (ss-path buffer-file-name)))
    (ss-enqueue 'ss-parse-diff "ss" nil "diff" "-I-" ss-diff-format-args (ss-path buffer-file-name)))
  (ss-run-tasks))

(defun ss-parse-diff (rc ss-output-buf)
  "Display results of ss-diff command"
  (cond ((equal rc 0)
	   (message "No differences"))
	  ((equal rc 1)
           (ss-display-output ss-output-buf "*Source Safe Diff Report*"))
	  (t
	   (error "Diff operation failed"))))



(defun ss-directory (dir)
  "Display Source Safe directory."
  (interactive "DProject: ")
  (if dir
      (progn
        (ss-enqueue 'ss-parse-directory "ss" nil "dir" "-E" "-R" "-I-" (ss-path dir))
        (ss-run-tasks))))

(defun ss-parse-directory (rc ss-output-buf)
  "Parse ss-directory output."
  (if (equal rc 0)
      (ss-display-output ss-output-buf "*Source Safe Directory*")
    (error "Directory operation failed.")))



(defun ss-files-checked-out-by-me (dir)
  "Display all files checked out by ss-user-name in a project and it's sub-projects."
  (interactive "DProject: ")
  (if dir
      (progn
        (ss-enqueue 'ss-parse-checked-out-by-me
                    "ss" nil "status" "-R" (concat "-U" (ss-user-name)) "-I-"
                    (ss-path dir))
        (ss-run-tasks))))

(defun ss-parse-checked-out-by-me (rc ss-output-buf)
  (cond ((equal rc 0)
         (message "No checked out files found."))
        ((equal rc 1)
         (ss-display-output ss-output-buf (concat "*Files Checked Out By " (ss-user-name) "*"))) 
        (t
         (error "Status operation failed."))))



(defun ss-undo-checkout ()
  "Undo checkout of file.
This only works for files that have not been modified."
  (interactive)
  (ss-buffer-file-check)
  (ss-enqueue 'ss-parse-undo "ss" (lambda () "Y\n") "undo" (ss-path buffer-file-name))
  (ss-run-tasks))

(defun ss-parse-undo (rc ss-output-buf)
  "Parse results of ss-undo-checkout."
  (if (equal rc 0)
      (progn
        (ss-internal-revert-buffer)
        (setq buffer-read-only t)
        (setq ss-status (list 'state 'checked-in))
        (ss-update-mode-line))
    (error "Undo checkout operation failed")))



(defun ss-db-menu-filter (items)
  "List ss database choices."
  (let ((choices ss-db-list)
	menus)
    (while choices
      (setq menus (append menus (list (vector (car (car choices))
					      (list 'setenv "SSDIR" (car (cdr (car choices))))
						    :active t))))
      (setq choices (cdr choices)))
    menus))
      
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;




;; Utilities ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun ss-user-name ()
  "Returns ss-user-name.
If variable ss-user-name is nil, set from environment variable USERNAME.
Use this function to query the value of the ss-user-name variable.
Don't access the ss-user-name variable directly."
  (or (and ss-user-name (> (length ss-user-name) 0))
      (setq ss-user-name (getenv "USERNAME")))
  ss-user-name)



(defun ss-path (path)
  "Changes path into Source Safe friendly path.
Returns the new string."
  (if (equal (aref path 1) ?:)
      (setq path (substring path 2)))
  (let ((i 0))
    (while (< i (length path))
      (if (char-equal (aref path i) ?\\)
	   (aset path i ?/))
      (setq i (1+ i))))
  (setq path (concat "$" path)))


(defun ss-buffer-file-check ()
  "Causes error if current buffer does not have an associated file."
  (if (not buffer-file-name)
      (error "Buffer has no associated file")))


(defvar ss-status-hook-on t
  "Controls executions of ss-status-hook")


(defun ss-status-hook ()
  "Call ss-status if it seems that current buffer file might be in Source Safe.
Also check the value of ss-status-hook-on."
  (if (and
	  ss-status-hook-on
	  buffer-file-name
	  (or (not ss-file-extensions-list)
	      (let ((pos (string-match "\\." buffer-file-name)))
		(and pos (member (substring buffer-file-name pos) ss-file-extensions-list))))
	  (or (not ss-directories-list)
	      (let ((dirs ss-directories-list)
		    (found nil))
		(while (and dirs (not found))
		  (setq found (string-match (car dirs) buffer-file-name))
		  (setq dirs (cdr dirs)))
		found)))
      (ss-status)))


(defun ss-internal-revert-buffer ()
  "Revert buffer in ss-mode.  Used interally by ss-mode functions."
  (setq ss-status-hook-on nil)
  (unwind-protect
      ; try to put point, mark in roughly the same positions in new file
      (let ((pnt (point))
	    (mrk (mark)))
	(revert-buffer t t)
	(goto-char pnt)
	(set-mark mrk))
    (setq ss-status-hook-on t)))


(defun ss-display-output (ss-output-buf name)
  "Display results of ss process output in new buffer."
  (let ((new-buf (generate-new-buffer name)))
    (save-excursion
      (set-buffer new-buf)
      (insert-buffer ss-output-buf))
    (display-buffer new-buf)))


(defun ss-message (format-str &rest args)
  (condition-case err
      (message format-str args)
    (error
     )))
      
  
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(provide 'ss-mode)

;;; ss-mode.el ends herex
